# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SVAMP word problem dataset."""
import json
import random
import re
from typing import Text, Optional, Union

import cascades as cc
from cascades.examples import calculator

open_file = open
SVAMP_PATH = None


def load_svamp(seed=0, valid_size=50, test_size=50, json_path=SVAMP_PATH):
  """Load SVAMP dataset from json.

  Available at:
    https://raw.githubusercontent.com/arkilpatel/SVAMP/main/SVAMP.json

  Args:
    seed: Seed for shuffling.
    valid_size: Number of examples in validation set.
    test_size: Number of examples in test set.
    json_path: Json file path to load.

  Returns:
    A dictionary of (train, valid, test).
  """

  if not json_path:
    raise ValueError('Missing json_path for load_svamp must be specified.')

  with open_file(json_path) as f:
    dataset = json.loads(f.read())
  random.Random(seed).shuffle(dataset)
  return dict(
      valid=dataset[:valid_size],
      test=dataset[valid_size:valid_size + test_size],
      train=dataset[valid_size + test_size:])


def _evaluate_equation(text: Text) -> Optional[float]:
  try:
    output = calculator.eval_arithmetic(text)
    return output
  except (AttributeError, SyntaxError, ZeroDivisionError) as e:
    del e
    return None


def _score_prediction(model_answer: Union[float, int, Text],
                      true_answer: Union[float, int, Text]):
  model_answer = str(model_answer)
  true_answer = str(true_answer)
  if true_answer == model_answer:
    return True
  else:
    return False


def _svamp_format_task(task):
  return '{body} : {question}'.format(
      body=task['Body'], question=task['Question'])


def _text_extract_regex(text: Text, pattern: Text) -> Optional[Text]:
  p = re.compile(pattern)
  matches = p.findall(text)
  if not matches:
    return None
  matched_text = matches[0]
  return matched_text


def _text_extract_brackets(text: Text) -> Optional[Text]:
  """Extract text within {brackets}."""
  return _text_extract_regex(text, pattern=r'{{(.*?)}}')


def _svamp_format_prompt(
    task,
    examples,
    instructions='Write the arithmetic expression to solve each word problem:'):
  """Create prompt based on svamp task."""
  formatted_io_examples = [(_svamp_format_task(task),
                            '{{ %s }}' % task['Equation']) for task in examples]
  parts = [instructions]
  for question, answer in formatted_io_examples:
    parts.append('QUESTION: %s\nANSWER: %s' % (question, answer))

  parts.append('QUESTION: %s\nANSWER:' % _svamp_format_task(task))
  return '\n\n'.join(parts)


def solve_svamp_task(task, train_set, lm=None, check_correct=False):
  """Generate solution questions to the SVAMP task, conditioned on train_set."""
  # TODO(ddohan): Make distribution over conditioning questions a part of model.
  prompt = _svamp_format_prompt(task=task, examples=train_set)

  # Generate equations
  line = yield cc.sample(
      cc.String(prompt=prompt, lm=lm, until='\n'), name='gen_equation')

  # Clip to 1 line and rescore
  # line = line.strip().split('\n')[0]

  # TODO(ddohan): Should this go in the prior p(x) instead of likelihood p(y|x)
  # yield cc.observe(obs=line, dist=lm(prompt=prompt), name='clip_and_rescore')

  # Extract equations between {{ }}
  eq = _text_extract_brackets(line)
  eq = eq.strip()
  if eq is None:
    yield cc.reject(reason='No equation proposed', name='no_equation')
  eq = eq.replace('{', '').replace('}', '')

  # Evaluate the arithmetic using a calculator
  calculator_output = _evaluate_equation(eq)

  if calculator_output is None:
    yield cc.reject(reason=f'Calculator failed: `{eq}`')

  if check_correct:
    # Did we get the right answer?
    correct = _score_prediction(model_answer=calculator_output,
                                true_answer=task['Answer'])
    yield cc.log(correct, 'is_correct')
    if not correct:
      yield cc.reject(
          reason=f'Incorrect answer: expected {task["Answer"]} != {eq} from {calculator_output}',
          name='incorrect')

  return (eq, calculator_output)


# Maximize p(score | task) marginalized over solutions
# params = set of prompts we use in the prefix.
def svamp_optimize_prompt(lm, task, train_set, k=5):
  """Choose k tasks to put into prompt from training examples."""
  idxs = yield cc.sample(dist=cc.Choose(k=k, options=range(len(train_set))))  # pytype: disable=wrong-arg-types
  prompt_tasks = [train_set[i] for i in idxs]
  result = yield solve_svamp_task(lm=lm, task=task, train_set=prompt_tasks)
  return result



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base types for inference."""

import dataclasses
from typing import Any, Dict, Text, Optional


@dataclasses.dataclass
class InferenceResult:
  metadata: Any = None
  output: Any = None
  trace: Any = None
  inputs: Any = None

  def __getitem__(self, k):
    return self.trace[k]


class InferenceAlgorithm:
  """Base class for inference methods."""

  def __init__(self,
               model,
               *args,
               observe: Optional[Dict[Text, Any]] = None,
               **kwargs):
    """Run inference on model called with given inputs.

    Args:
      model: Model to run sampling on.
      *args: Arguments passed to the model.
      observe: Values to condition the model on.
      **kwargs: Keyword args passed to the model.

    Returns:
      A sample from the model.
    """
    self.model = model
    self.args = args
    self.kwargs = kwargs
    self.observe = observe

  def sample(self):
    raise NotImplementedError



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Scratchpad/chain of thought, demonstrated on gsm8k.

Scratchpad: https://arxiv.org/abs/2112.00114
Chain of thought: https://arxiv.org/abs/2201.11903

Self consistency (https://arxiv.org/abs/2203.11171) is done by marginalizing out
  the `thought` in the `infer` operator.
"""
import dataclasses
import re
from typing import Dict, Iterable, Tuple, Optional, Text, List
import cascades as cc
from cascades.examples.tasks import gsm8k

# Standard chain of thought arithmetic prompts.
# From https://arxiv.org/abs/2112.00114
PROMPTS = """Q: There are 15 trees in the grove. Grove workers will plant trees in the grove today. After they are done, there will be 21 trees. How many trees did the grove workers plant today?
A: We start with 15 trees. Later we have 21 trees. The difference must be the number of trees they planted. So, they must have planted 21 - 15 = 6 trees. The answer is 6.
Q: If there are 3 cars in the parking lot and 2 more cars arrive, how many cars are in the parking lot?
A: There are 3 cars in the parking lot already. 2 more arrive. Now there are 3 + 2 = 5 cars. The answer is 5.
Q: Leah had 32 chocolates and her sister had 42. If they ate 35, how many pieces do they have left in total?
A: Leah had 32 chocolates and Leah’s sister had 42. That means there were originally 32 + 42 = 74 chocolates. 35 have been eaten. So in total they still have 74 - 35 = 39 chocolates. The answer is 39.
Q: Jason had 20 lollipops. He gave Denny some lollipops. Now Jason has 12 lollipops. How many lollipops did Jason give to Denny?
A: Jason had 20 lollipops. Since he only has 12 now, he must have given the rest to Denny. The number of lollipops he has given to Denny must have been 20 - 12 = 8 lollipops. The answer is 8.
Q: Shawn has five toys. For Christmas, he got two toys each from his mom and dad. How many toys does he have now?
A: He has 5 toys. He got 2 from mom, so after that he has 5 + 2 = 7 toys. Then he got 2 more from dad, so in total he has 7 + 2 = 9 toys. The answer is 9.
Q: There were nine computers in the server room. Five more computers were installed each day, from monday to thursday. How many computers are now in the server room?
A: There are 4 days from monday to thursday. 5 computers were added each day. That means in total 4 * 5 = 20 computers were added. There were 9 computers in the beginning, so now there are 9 + 20 = 29 computers. The answer is 29.
Q: Michael had 58 golf balls. On tuesday, he lost 23 golf balls. On wednesday, he lost 2 more. How many golf balls did he have at the end of wednesday?
A: Michael initially had 58 balls. He lost 23 on Tuesday, so after that he has 58 - 23 = 35 balls. On Wednesday he lost 2 more so now he has 35 - 2 = 33 balls. The answer is 33.
Q: Olivia has $23. She bought five bagels for $3 each. How much money does she have left?
A: She bought 5 bagels for $3 each. This means she spent 5 * $3 = $15 on the bagels. She had $23 in beginning, so now she has $23 - $15 = $8. The answer is 8.
"""


@dataclasses.dataclass
class ReasonIO:
  """Represents a question and answer, together with reasoning."""
  question: Text
  reason: Optional[Text] = None
  answer: Optional[Text] = None
  id: Optional[Text] = dataclasses.field(default=None)


# From https://arxiv.org/abs/2203.11171
def _process_part(qa: Text) -> ReasonIO:
  question, reason = qa.split('A:')
  question = question.replace('Q:', '')
  reason, answer = reason.split('The answer is')
  answer = answer.replace('.', '')
  return ReasonIO(
      question=question.strip(), reason=reason.strip(), answer=answer.strip())


def load_chain_examples() -> Tuple[ReasonIO]:
  """Load the standard chain of thought prompts used in the paper."""
  parts = PROMPTS.split('Q:')[1:]
  chain_prompts = tuple(_process_part(x) for x in parts)
  return chain_prompts


def load_gsm8k(base_dir=gsm8k.GSM8K_PATH):
  """Load and process the gsm8k dataset."""
  splits = gsm8k.load_dataset(base_dir=base_dir)
  train = tuple(map(process_gsm8k_example, splits['train']))
  test = tuple(map(process_gsm8k_example, splits['test']))
  return dict(train=train, test=test)


def process_gsm8k_example(x: Dict[Text, Text]) -> ReasonIO:
  """Convert gsm8k dicts into ReasonIO."""
  return ReasonIO(
      question=x['question'],
      reason=x['answer'].split('####')[0].strip(),
      answer=x['final_answer'],
      id=x['id'])


def fewshot_prompt(examples: Iterable[ReasonIO],
                   target: Optional[ReasonIO] = None):
  """Construct a few shot prompt."""
  parts = []
  for x in examples:
    if x.reason is None:
      raise ValueError('Must provide reason for few shot.')
    if x.answer is None:
      raise ValueError('Must provide answer for few shot.')
    part = f'Question: {x.question}\nReason: {x.reason}\nAnswer: {x.answer}'
    parts.append(part)
  if target:
    part = f'Question: {target.question}\nReason'
    parts.append(part)
  else:
    parts.append('')

  prompt = '\n===\n'.join(parts).strip()
  return prompt


def sample_with_prompts(target: ReasonIO,
                        examples: List[ReasonIO],
                        n_prompts: Optional[int] = None,
                        lm=None,
                        max_calls=4):
  """Sample an answer for target given prompts.

  Args:
    target: Target task.
    examples: Prompting tasks.
    n_prompts: If None, use all examples in prompt. Otherwise, select this many
      examples to place in prompt.
    lm: Language model to use.
    max_calls: Max # of times to iterate the LM sampling.

  Yields:
    Cascade distributions (all LM sample nodes in this case).

   Returns:
     predicted answer string
  """
  # Select a random subset and ordering of the prompts.
  yield cc.log(value=target.id, name='problem_id')
  if n_prompts:
    chosen_examples = yield cc.sample(
        cc.Choose(k=n_prompts, options=examples), name='choose_prompts')
  else:
    chosen_examples = examples

  # Create few shot prompt
  prompt = fewshot_prompt(examples=chosen_examples, target=target)

  # Sample until we hit the end of example marker (`===`), then extract
  # the answer as rightmost digits.
  prediction = yield cc.sample(
      cc.String(prompt=prompt, lm=lm, k=max_calls, until='==='), name='thought')

  # Find right most number.
  nums = re.findall(r'\d+', prediction)
  if nums:
    answer = nums[-1]
  else:
    yield cc.reject(reason=f'No answer found in `{prediction}`')
  return answer



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic tests for Scratchpads."""
import functools

from absl.testing import absltest
import cascades as cc
from cascades.examples import scratchpad


class ScratchpadTest(absltest.TestCase):

  def test_sample_solution(self):
    examples = scratchpad.load_chain_examples()
    target = scratchpad.ReasonIO(
        question='Natalia sold clips to 48 of her friends in April, and then she sold half as many clips in May. How many clips did Natalia sell altogether in April and May?',
        reason=None,
        answer='72',
        id='test/123')

    mock_lm = cc.mock_lm(
        response=': Half of 48 is 24. 24 + 48 is 72.\nAnswer: 72\n===')

    program = functools.partial(
        scratchpad.sample_with_prompts,
        lm=mock_lm,
        target=target,
        examples=examples,
        n_prompts=3)
    sampler = cc.Sampler(model=program)  # pylint: disable=unnecessary-lambda
    trace = sampler.reify(seed=0)
    self.assertEqual('72', trace.return_value)
    self.assertEqual('test/123', trace['problem_id'].value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for rejection_sampling."""
from absl.testing import absltest
import cascades as cc
from cascades._src.inference import rejection_sampling
from numpyro import distributions as np_dists


@cc.model
def truncated_2step_gaussian(cutoff):
  """Two step random walk with rejection criteria."""
  step1 = yield cc.sample(name='step 1', dist=np_dists.Normal(loc=0, scale=1))
  step2 = yield cc.sample(name='step 2', dist=np_dists.Normal(loc=0, scale=1))
  output = step1 + step2
  if abs(output) >= cutoff:
    yield cc.reject('Out of bounds')
  return output


class RejectionSamplingTest(absltest.TestCase):

  def test_sample_with_reject(self):
    """Check sampling until a trace with a finite likelihood."""
    cutoff = 0.3
    s = rejection_sampling.RejectionSampling(
        model=truncated_2step_gaussian,
        max_attempts=100,
        cutoff=cutoff)
    for seed in range(10):
      y = s.sample(seed=seed)
      self.assertLess(abs(y.output), cutoff)

  def test_sample_with_reject_conditioned(self):
    """Check that we can condition specific values."""
    cutoff = 0.5
    s = rejection_sampling.RejectionSampling(
        model=truncated_2step_gaussian, max_attempts=100,
        observe={'step 1': 2.0},
        cutoff=cutoff)
    # Jumps to 2 then to < 0.5
    y = s.sample(seed=0)
    self.assertLess(y.output, cutoff)
    self.assertEqual(2.0, y.trace['step 1'].value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GSM8k reasoning task."""
import json
import os

open_file = open
GSM8K_PATH = None


def _load_line(line):
  example = json.loads(line)
  answer = int(example['answer'].split('#### ')[-1].replace(',', ''))
  example['final_answer'] = answer
  return example


def load_dataset(base_dir=GSM8K_PATH):
  """Load GSM8k raw data from `base_dir`.

  Available at:
    https://raw.githubusercontent.com/arkilpatel/SVAMP/main/SVAMP.json

  Args:
    base_dir: Directory containing `train.jsonl` and `test.jsonl`.

  Returns:
    A dictionary of (train, test).
  """
  train_path = os.path.join(base_dir, 'train.jsonl')
  test_path = os.path.join(base_dir, 'test.jsonl')

  train_xs = []
  test_xs = []

  with open_file(train_path) as f:
    for i, line in enumerate(f):
      example = _load_line(line)
      example['id'] = f'train/{i}'
      train_xs.append(example)

  with open_file(test_path) as f:
    for i, line in enumerate(f):
      example = _load_line(line)
      example['id'] = f'test/{i}'
      test_xs.append(example)

  return dict(train=train_xs, test=test_xs)



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Rejection sampling from a model."""

import math
from typing import Any, Dict, Optional

from cascades._src.distributions import base as dists
from cascades._src.inference import base
import jax


class RejectionSampling(base.InferenceAlgorithm):
  """Draw samples via rejection sampling."""

  def __init__(self,
               model,
               *args,
               observe: Optional[Dict[str, Any]] = None,
               max_attempts=20,
               **kwargs):
    """Instantiate a rejection sampler for given model.

    Args:
      model: Model to run sampling on.
      *args: Arguments passed to the model.
      observe: Values to condition the model on.
      max_attempts: Maximum samples to attempt.
      **kwargs: Keyword args passed to the model.

    Returns:
      A sample from the model.
    """
    super().__init__(model=model, *args, observe=observe, **kwargs)
    self.max_attempts = max_attempts

  def sample(self, seed=0):
    """Generate a sample via rejection sampling."""
    output = None
    rng = dists.get_rng(seed)
    for i in range(self.max_attempts):
      rng, subrng = jax.random.split(rng)
      trace = self.model.sample(
          *self.args, seed=subrng, observe=self.observe, **self.kwargs)
      if not math.isinf(trace.observed_likelihood):
        output = trace.return_value
        break
    return base.InferenceResult(
        metadata=dict(attempts=i + 1),
        trace=trace,
        inputs=(self.args, self.kwargs),
        output=output)



# Copyright 2022 The cascades Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for reasoning with a calculator."""
import ast
import operator as op

# https://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string
# supported operators
operators = {
    ast.Add: op.add,
    ast.Sub: op.sub,
    ast.Mult: op.mul,
    ast.Div: op.truediv,
    ast.Pow: op.pow,
    ast.BitXor: op.xor,
    ast.USub: op.neg
}


def eval_arithmetic(expr):
  """Eval an arithmetic expression.

    >>> eval_expr('2^6')
    4
    >>> eval_expr('2**6')
    64
    >>> eval_expr('1 + 2*3**(4^5) / (6 + -7)')
    -5.0

  Args:
    expr: Expression to evaluate

  Returns:
    Result of evaluating expression.
  """
  return _eval_arithmetic(ast.parse(expr, mode='eval').body)


def _eval_arithmetic(node):
  if isinstance(node, ast.Num):  # <number>
    return node.n
  elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
    return operators[type(node.op)](_eval_arithmetic(node.left),
                                    _eval_arithmetic(node.right))
  elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
    return operators[type(node.op)](_eval_arithmetic(node.operand))
  else:
    raise TypeError(node)
